@@ -30,6 +30,7 @@ import sys
 
 import cv2
 import matplotlib
+matplotlib.use("Qt5Agg")
 import matplotlib.patches as patches
 import matplotlib.pyplot as plt
 import numpy as np
@@ -37,13 +38,12 @@ from matplotlib.path import Path
 from scipy.optimize import linear_sum_assignment
 from scipy.spatial import distance
 from scipy.spatial.distance import cdist
+from scipy.stats import kde
 from sklearn.cluster import KMeans
+import pandas as pd
 
 from doris import config
 
-matplotlib.use("Qt5Agg")
-
-
 
 def rgbstr_to_bgr_tuple(rgb_str):
     return struct.unpack('BBB', bytes.fromhex(rgb_str))[::-1]
@@ -56,13 +56,28 @@ def plot_path(df, x_lim, y_lim):
     plot path
     """
 
+    print("df", df)
+
     plt.figure()
     axes = plt.gca()
     axes.set_aspect("equal", adjustable="box")
 
     for idx in range(1, int((len(df.columns) - 1)/2) + 1):
         plot_color = "#" + config.RGBSTR_COLORS_LIST[(idx - 1) % len(config.RGBSTR_COLORS_LIST)]
-        plt.plot(f"x{idx}", f"y{idx}", plot_color, data=df, alpha=1)
+        
+        plt.plot(df[f"x{idx}"].to_numpy(float, na_value=np.nan),
+                 df[f"y{idx}"].to_numpy(float, na_value=np.nan),
+                 color=plot_color,
+                 alpha=1)
+        
+
+        '''
+        plt.plot(df[f"x{idx}"],
+                 df[f"y{idx}"],
+                 color=plot_color,
+                 alpha=1)
+        '''
+
 
     plt.xlabel("x")
     plt.ylabel("y")
@@ -87,13 +102,18 @@ def plot_positions(df, x_lim, y_lim,):
 
     #plt.figure(figsize=(5,5))
 
+    print("df", df)
+
     plt.figure()
     axes = plt.gca()
     axes.set_aspect("equal", adjustable="box")
 
     for idx in range(1, int((len(df.columns) - 1)/2) +1):
         plot_color = "#" + config.RGBSTR_COLORS_LIST[(idx - 1) % len(config.RGBSTR_COLORS_LIST)]
-        plt.scatter(df[f"x{idx}"], df[f"y{idx}"], c=plot_color, alpha=0.5)
+        plt.scatter(df[f"x{idx}"].to_numpy(float, na_value=np.nan),
+                    df[f"y{idx}"].to_numpy(float, na_value=np.nan),
+                    c=plot_color,
+                    alpha=0.5)
 
     plt.xlabel('x')
     plt.ylabel('y')
@@ -106,38 +126,28 @@ def plot_positions(df, x_lim, y_lim,):
     plt.show()
 
 
-def plot_density_old(x, y, x_lim=(0, 0), y_lim=(0,0)):
-
-    try:
-        x = np.array(x)
-        y = np.array(y)
-
-        fig = plt.figure()
-        ax = fig.add_subplot(1, 1, 1, projection="scatter_density")
-        ax.scatter_density(x, y)
-        if x_lim != (0, 0):
-            ax.set_xlim(x_lim)
-        if y_lim != (0, 0):
-            ax.set_ylim(y_lim[::-1])
-
-        plt.show()
-        return True
-
-    except Exception:
-        return False
-
-
 def plot_density(df, x_lim, y_lim):
 
-    df = df.dropna(thresh=1)
+    nbins=100
+    i = 1
+    while True:
+    
+        if f"x{i}" not in df:
+            return
+        
+        df[f"x{i}"] = pd.to_numeric(df[f"x{i}"], errors='coerce')
+        df[f"y{i}"] = pd.to_numeric(df[f"y{i}"], errors='coerce')
+
+        k = kde.gaussian_kde((df[f"x{i}"].dropna(), df[f"y{i}"].dropna()))
+
+        xi, yi = np.mgrid[df[f"x{i}"].dropna().min():df[f"x{i}"].dropna().max():nbins*1j,
+                          df[f"y{i}"].dropna().min():df[f"y{i}"].dropna().max():nbins*1j]
+        zi = k(np.vstack([xi.flatten(), yi.flatten()]))
 
-    for idx in range(1, int((len(df.columns) - 1) / 2) + 1):
         plt.figure()
         axes = plt.gca()
         axes.set_aspect("equal", adjustable="box")
 
-        plt.hist2d(df[f"x{idx}"], df[f"y{idx}"], bins=20, cmap=plt.cm.Reds)
-
         plt.xlabel("x")
         plt.ylabel("y")
 
@@ -145,8 +155,13 @@ def plot_density(df, x_lim, y_lim):
         axes.set_ylim(y_lim)
         axes.set_ylim(axes.get_ylim()[::-1])
 
+        plt.title(f"Density plot for object #{i}")
+        plt.pcolormesh(xi, yi, zi.reshape(xi.shape))  # , cmap=plt.cm.BuGn_r
+        plt.colorbar()
         plt.tight_layout()
         plt.show()
+        i += 1
+
 
 
 def group_of(points, centroids):
@@ -449,9 +464,10 @@ def find_circle(points):
         points (list): list of points
 
     Returns:
-        float: x of circle center
-        float: y of circle center
-        float: radius of circle
+        float: x of circle center or 0 in case of error
+        float: y of circle center or 0 in case of error
+        float: radius of circle or -1 in case of error
+
     """
 
     x1, y1 = points[0]
@@ -466,11 +482,14 @@ def find_circle(points):
 
     mb = (y3 - y2) / (x3 - x2)
 
-    x = (ma * mb * (y1 - y3) + mb * (x1 + x2) - ma * (x2 + x3)) / (2 * (mb - ma))
+    try:
+        x = (ma * mb * (y1 - y3) + mb * (x1 + x2) - ma * (x2 + x3)) / (2 * (mb - ma))
+        y = - (1 / ma) * (x - (x1 + x2) / 2) + (y1 + y2) / 2
+    except ZeroDivisionError:
+        return (0, 0, -1)
 
-    y = - (1 / ma) * (x - (x1 + x2) / 2) + (y1 + y2) / 2
 
-    return x, y, euclidean_distance((x, y), (x1, y1))
+    return (x, y, euclidean_distance((x, y), (x1, y1)))
 
 
 def detect_and_filter_objects(frame,
@@ -487,7 +506,7 @@ def detect_and_filter_objects(frame,
         max_size (int): maximum size of objects to detect
         number_of_objects (int): number of objects to return
         arena (list): list of arena
-        max_extension (int): maximum extension of object to select
+        max_extension (int): maximum extension of object to select (in pixels)
 
     Returns:
         dict: all detected objects
@@ -501,19 +520,20 @@ def detect_and_filter_objects(frame,
 
     all_objects = {}
     for idx, cnt in enumerate(contours):
-        M = cv2.moments(cnt)
-        if M["m00"] != 0:
-            cx = int(M["m10"] / M["m00"])
-            cy = int(M["m01"] / M["m00"])
-        else:
-            cx, cy = 0, 0
 
         n = np.vstack(cnt).squeeze()
         try:
             x, y = n[:,0], n[:,1]
         except:
-            x = n[0]
-            y = n[1]
+            x, y = n[0], n[1]
+
+        # centroid
+        M = cv2.moments(cnt)
+        if M["m00"] != 0:
+            cx = int(M["m10"] / M["m00"])
+            cy = int(M["m01"] / M["m00"])
+        else:
+            cx, cy = np.mean(x), np.mean(y)
 
         all_objects[idx] = {"centroid": (cx, cy),
                             "contour": cnt,
